# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import uuid

import pytest

import argilla as rg
from argilla.records._resource import Record


@pytest.fixture
def record():
    return rg.Record(
        id=uuid.uuid4(),
        fields={"text": "Hello World, how are you?"},
        suggestions=[
            rg.Suggestion("label", "positive", score=0.9),
            rg.Suggestion("topics", ["topic1", "topic2"], score=[0.9, 0.8]),
        ],
        responses=[rg.Response("label", "positive", user_id=uuid.uuid4())],
        metadata={"source": "twitter", "language": "en"},
        vectors={"text": [0, 0, 0]},
    )


def test_export_record_to_from_dict(record):
    record_dict = record.to_dict()
    imported_record = rg.Record.from_dict(record_dict)

    assert record.responses["label"][0].value == imported_record.responses["label"][0].value
    assert record.suggestions["topics"].value == imported_record.suggestions["topics"].value
    for key, value in record.metadata.items():
        assert imported_record.metadata[key] == value
    assert record.fields["text"] == imported_record.fields["text"]
    # This is a consequence of how UUIDs are treated in python and could be
    #  problematic for users.
    assert str(record.id) == imported_record.id


def test_export_generic_io_via_json(record):
    record_dict = record.to_dict()
    record_dict = json.dumps(record_dict)
    record_dict = json.loads(record_dict)
    imported_record = Record.from_dict(record_dict)

    assert record.responses["label"][0].value == imported_record.responses["label"][0].value
    assert record.suggestions["topics"].value == imported_record.suggestions["topics"].value
    for key, value in record.metadata.items():
        assert imported_record.metadata[key] == value
    assert record.fields["text"] == imported_record.fields["text"]
    assert record.vectors["text"] == imported_record.vectors["text"]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Optional, Union

from argilla._models import FieldModel, QuestionBaseModel
from argilla._resource import Resource

__all__ = ["SettingsPropertyBase"]


class SettingsPropertyBase(Resource):
    """Base class for dataset fields or questions in Settings class"""

    _model: Union[FieldModel, QuestionBaseModel]

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={self.name}, title={self.title}, description={self.description}, "
            f"type={self.type}, required={self.required}) \n"
        )

    @property
    def name(self) -> str:
        return self._model.name

    @property
    def title(self) -> Optional[str]:
        return self._model.title

    @property
    def required(self) -> bool:
        return self._model.required

    @property
    def description(self) -> Optional[str]:
        return self._model.description

    @property
    def type(self) -> str:
        return self._model.settings.type

    def serialize(self) -> dict[str, Any]:
        serialized_model = super().serialize()
        serialized_model["type"] = self.type
        return serialized_model



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4

import argilla as rg
from argilla.records._io import GenericIO


class TestGenericIO:
    def test_to_list_flatten(self):
        user_a, user_b, user_c = uuid4(), uuid4(), uuid4()

        record = rg.Record(
            fields={"field": "The field"},
            metadata={"key": "value"},
            responses=[
                rg.Response(question_name="q1", value="value", user_id=user_a),
                rg.Response(question_name="q2", value="value", user_id=user_a),
                rg.Response(question_name="q2", value="value", user_id=user_b),
                rg.Response(question_name="q1", value="value", user_id=user_c),
            ],
            suggestions=[
                rg.Suggestion(question_name="q1", value="value", score=0.1, agent="test"),
                rg.Suggestion(question_name="q2", value="value", score=0.9),
            ],
        )

        records_list = GenericIO.to_list([record], flatten=True)
        assert records_list == [
            {
                "id": str(record.id),
                "_server_id": None,
                "field": "The field",
                "key": "value",
                "q1.responses": ["value", "value"],
                "q1.responses.users": [str(user_a), str(user_c)],
                "q2.responses": ["value", "value"],
                "q2.responses.users": [str(user_a), str(user_b)],
                "q1.suggestion": "value",
                "q1.suggestion.score": 0.1,
                "q1.suggestion.agent": "test",
                "q2.suggestion": "value",
                "q2.suggestion.score": 0.9,
                "q2.suggestion.agent": None,
            }
        ]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import json
from pathlib import Path
from typing import List, Union

from argilla.records._resource import Record
from argilla.records._io import GenericIO


class JsonIO:
    @staticmethod
    def to_json(records: List["Record"], path: Union[Path, str]) -> Path:
        """
        Export the records to a file on disk. This is a convenient shortcut for dataset.records(...).to_disk().

        Parameters:
            path (str): The path to the file to save the records.
            orient (str): The structure of the exported dictionary.

        Returns:
            The path to the file where the records were saved.

        """
        if isinstance(path, str):
            path = Path(path)
        if path.exists():
            raise FileExistsError(f"File {path} already exists.")
        record_dicts = GenericIO.to_list(records, flatten=False)
        with open(path, "w") as f:
            json.dump(record_dicts, f)
        return path

    @staticmethod
    def _records_from_json(path: Union[Path, str]) -> List["Record"]:
        """Creates a DatasetRecords object from a disk path.

        Parameters:
            path (str): The path to the file containing the records.

        Returns:
            DatasetRecords: The DatasetRecords object created from the disk path.

        """
        with open(path, "r") as f:
            record_dicts = json.load(f)
        records = [Record.from_dict(record_dict) for record_dict in record_dicts]
        return records



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import pytest

from argilla import Dataset, Settings, TextField, LabelQuestion, Argilla, VectorField, FloatMetadataProperty


@pytest.fixture
def dataset():
    return Dataset(
        name=f"test_dataset_{uuid.uuid4().int}",
        settings=Settings(
            fields=[TextField(name="text", use_markdown=False)],
            questions=[LabelQuestion(name="label", labels=["a", "b", "c"])],
        ),
    ).create()


class TestUpdateDatasetSettings:
    def test_update_settings(self, client: Argilla, dataset: Dataset):
        settings = dataset.settings

        settings.fields.text.use_markdown = True
        dataset.settings.vectors.add(VectorField(name="vector", dimensions=10))
        dataset.settings.metadata.add(FloatMetadataProperty(name="metadata"))
        dataset.settings.update()

        dataset = client.datasets(dataset.name)
        settings = dataset.settings
        assert settings.fields["text"].use_markdown is True
        assert settings.vectors["vector"].dimensions == 10
        assert isinstance(settings.metadata.metadata, FloatMetadataProperty)

        settings.vectors["vector"].title = "A new title for vector"

        settings.update()
        dataset = client.datasets(dataset.name)
        assert dataset.settings.vectors["vector"].title == "A new title for vector"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import List, Dict
from uuid import UUID

import httpx
from argilla._api._base import ResourceAPI
from argilla._exceptions import api_error_handler
from argilla._models import (
    TextQuestionModel,
    LabelQuestionModel,
    MultiLabelQuestionModel,
    RankingQuestionModel,
    RatingQuestionModel,
    SpanQuestionModel,
    QuestionBaseModel,
    QuestionModel,
)

__all__ = ["QuestionsAPI"]


class QuestionsAPI(ResourceAPI[QuestionBaseModel]):
    """Manage datasets via the API"""

    http_client: httpx.Client

    _TYPE_TO_MODEL_CLASS = {
        "text": TextQuestionModel,
        "label_selection": LabelQuestionModel,
        "multi_label_selection": MultiLabelQuestionModel,
        "ranking": RankingQuestionModel,
        "rating": RatingQuestionModel,
        "span": SpanQuestionModel,
    }

    ################
    # CRUD methods #
    ################

    @api_error_handler
    def create(
        self,
        dataset_id: UUID,
        question: QuestionModel,
    ) -> QuestionModel:
        url = f"/api/v1/datasets/{dataset_id}/questions"
        response = self.http_client.post(url=url, json=question.model_dump())
        response.raise_for_status()
        response_json = response.json()
        question_model = self._model_from_json(response_json=response_json)
        self._log_message(message=f"Created question {question_model.name} in dataset {dataset_id}")
        return question_model

    @api_error_handler
    def update(
        self,
        question: QuestionModel,
    ) -> QuestionModel:
        # TODO: Implement update method for fields with server side ID
        raise NotImplementedError

    @api_error_handler
    def delete(self, question_id: UUID) -> None:
        # TODO: Implement delete method for fields with server side ID
        raise NotImplementedError

    ####################
    # Utility methods #
    ####################

    def create_many(self, dataset_id: UUID, questions: List[QuestionModel]) -> List[QuestionModel]:
        response_models = []
        for question in questions:
            response_model = self.create(dataset_id=dataset_id, question=question)
            response_models.append(response_model)
        return response_models

    @api_error_handler
    def list(self, dataset_id: UUID) -> List[QuestionModel]:
        response = self.http_client.get(f"/api/v1/datasets/{dataset_id}/questions")
        response.raise_for_status()
        response_json = response.json()
        response_models = self._model_from_jsons(response_jsons=response_json["items"])
        return response_models

    ####################
    # Private methods #
    ####################

    def _model_from_json(self, response_json: Dict) -> QuestionModel:
        response_json["inserted_at"] = self._date_from_iso_format(date=response_json["inserted_at"])
        response_json["updated_at"] = self._date_from_iso_format(date=response_json["updated_at"])
        return self._get_model_from_response(response_json=response_json)

    def _model_from_jsons(self, response_jsons: List[Dict]) -> List[QuestionModel]:
        return list(map(self._model_from_json, response_jsons))

    def _get_model_from_response(self, response_json: Dict) -> QuestionModel:
        """Get the model from the response"""
        try:
            question_type = response_json.get("settings", {}).get("type")
        except Exception as e:
            raise ValueError("Invalid field type: missing 'settings.type' in response") from e

        question_class = self._TYPE_TO_MODEL_CLASS.get(question_type)
        if question_class is None:
            self._log_message(message=f"Unknown question type: {question_type}")
            question_class = QuestionBaseModel

        return question_class(**response_json, check_fields=False)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
from typing import Any, Dict, List, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from argilla import Record


class GenericIO:
    """This is a mixin class for DatasetRecords and Export classes.
    It handles methods for exporting records to generic python formats."""

    @staticmethod
    def to_list(records: List["Record"], flatten: bool = False) -> List[Dict[str, Union[str, float, int, list]]]:
        """Export records to a list of dictionaries with either names or record index as keys.
        Args:
            flatten (bool): The structure of the exported dictionary.
                - True: The record fields, metadata, suggestions and responses will be flattened.
                - False: The record fields, metadata, suggestions and responses will be nested.
        Returns:
            dataset_records (List[Dict[str, Union[str, float, int, list]]]): The exported records in a list of dictionaries format.
        """
        records_schema = set()
        dataset_records: list = []
        for record in records:
            record_dict = GenericIO._record_to_dict(record=record, flatten=flatten)
            records_schema.update([k for k in record_dict])
            dataset_records.append(record_dict)

        # normalize records structure
        for record_dict in dataset_records:
            record_dict.update({k: None for k in records_schema if k not in record_dict})

        return dataset_records

    @classmethod
    def to_dict(
        cls, records: List["Record"], flatten: bool = False, orient: str = "names"
    ) -> Dict[str, Union[str, float, int, list]]:
        """Export records to a dictionary with either names or record index as keys.
        Args:
            flatten (bool): The structure of the exported dictionary.
                - True: The record fields, metadata, suggestions and responses will be flattened.
                - False: The record fields, metadata, suggestions and responses will be nested.
            orient (str): The orientation of the exported dictionary.
                - "names": The keys of the dictionary will be the names of the fields, metadata, suggestions and responses.
                - "index": The keys of the dictionary will be the id of the records.
        Returns:
            dataset_records (Dict[str, Union[str, float, int, list]]): The exported records in a dictionary format.
        """
        if orient == "names":
            dataset_records: dict = defaultdict(list)
            for record in cls.to_list(records, flatten=flatten):
                for key, value in record.items():
                    dataset_records[key].append(value)
        elif orient == "index":
            dataset_records: dict = {}
            for record in records:
                dataset_records[record.id] = GenericIO._record_to_dict(record=record, flatten=flatten)
        else:
            raise ValueError(f"Invalid value for orient parameter: {orient}")
        return dict(dataset_records)

    ############################
    # Private methods
    ############################

    @staticmethod
    def _record_to_dict(record: "Record", flatten=False) -> Dict[str, Any]:
        """Converts a Record object to a dictionary for export.
        Args:
            record (Record): The Record object to convert.
            flatten (bool): The structure of the exported dictionary.
                - True: The record fields, metadata, suggestions and responses will be flattened
                        so that their keys becomes the keys of the record dictionary, using
                        dot notation for nested keys. i.e. `label.suggestion` and `label.response`
                - False: The record fields, metadata, suggestions and responses will be nested as
                        dictionaries within the record dictionary. i.e. `label: {suggestion: ..., response: ...}`
        Returns:
            A dictionary representing the record.
        """

        record_dict = record.to_dict()
        if flatten:
            record_dict.update(
                **record_dict.pop("fields", {}),
                **record_dict.pop("metadata", {}),
                **record_dict.pop("vectors", {}),
            )

            record_dict.pop("responses")
            record_dict.pop("suggestions")

            responses_dict = defaultdict(list)
            for response in record.responses:
                responses_key = f"{response.question_name}.responses"
                responses_users_key = f"{responses_key}.users"

                responses_dict[responses_key].append(response.value)
                responses_dict[responses_users_key].append(str(response.user_id))

            suggestions_dict = {}
            for suggestion in record.suggestions:
                suggestion_key = f"{suggestion.question_name}.suggestion"
                suggestion_agent_key = f"{suggestion_key}.agent"
                suggestion_score_key = f"{suggestion_key}.score"

                suggestions_dict.update(
                    {
                        suggestion_key: suggestion.value,
                        suggestion_score_key: suggestion.score,
                        suggestion_agent_key: suggestion.agent,
                    }
                )

            record_dict.update({**responses_dict, **suggestions_dict})
        return record_dict



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


# @pytest.fixture(scope="function", autouse=True)
# def mock_httpx_client(mocker) -> Generator[httpx.Client, None, None]:
#     mock_client = mocker.Mock(httpx.Client)
#     argilla.DEFAULT_HTTP_CLIENT = mock_client

#     return mock_client



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4

import pytest

import argilla as rg


@pytest.fixture
def dataset():
    settings = rg.Settings(
        fields=[rg.TextField(name="prompt")],
        questions=[rg.LabelQuestion(name="label", labels=["negative", "positive"])],
        metadata=[rg.FloatMetadataProperty(name="score")],
        vectors=[rg.VectorField(name="vector", dimensions=3)],
    )
    workspace = rg.Workspace(name="workspace", id=uuid4())
    return rg.Dataset(
        name="test_dataset",
        settings=settings,
        workspace=workspace,
    )


def test_ingest_record_from_dict(dataset):
    record_api_models = dataset.records._ingest_records(
        records=[
            {
                "prompt": "What is the capital of France?",
                "label": "positive",
            }
        ],
    )

    record = record_api_models[0]
    assert record.fields["prompt"] == "What is the capital of France?"
    assert record.suggestions[0].value == "positive"


def test_ingest_record_from_dict_with_mapped_suggestions(dataset):
    mock_mapping = {
        "my_prompt": "prompt",
        "my_label": "label.suggestion.value",
        "score": "label.suggestion.score",
        "model": "label.suggestion.agent",
    }
    record_api_models = dataset.records._ingest_records(
        records=[
            {
                "my_prompt": "What is the capital of France?",
                "my_label": "positive",
                "score": 0.9,
                "model": "model_name",
            }
        ],
        mapping=mock_mapping,
    )
    record = record_api_models[0]
    assert record.fields["prompt"] == "What is the capital of France?"
    assert record.suggestions[0].value == "positive"
    assert record.suggestions[0].question_name == "label"
    assert record.suggestions[0].score == 0.9
    assert record.suggestions[0].agent == "model_name"


def test_ingest_record_from_dict_with_mapped_responses(dataset):
    user_id = uuid4()
    mocked_mapping = {
        "label": "label.response",
    }
    record_api_models = dataset.records._ingest_records(
        records=[
            {
                "prompt": "Hello World, how are you?",
                "label": "negative",
            }
        ],
        mapping=mocked_mapping,
        user_id=user_id,
    )
    record = record_api_models[0]

    assert record.fields["prompt"] == "Hello World, how are you?"
    assert record.responses[0].values["label"]["value"] == "negative"
    assert record.responses[0].user_id == user_id


def test_ingest_record_from_dict_with_id_as_id(dataset):
    record_id = uuid4()
    record_api_models = dataset.records._ingest_records(
        records=[
            {
                "prompt": "Hello World, how are you?",
                "label": "negative",
                "id": record_id,
            }
        ],
    )
    record = record_api_models[0]
    assert record.fields["prompt"] == "Hello World, how are you?"
    assert record.external_id == record_id


def test_ingest_record_from_dict_with_mapped_id(dataset):
    record_id = uuid4()
    mock_mapping = {
        "test_id": "id",
    }
    record_api_models = dataset.records._ingest_records(
        records=[
            {
                "prompt": "Hello World, how are you?",
                "label": "negative",
                "test_id": record_id,
            }
        ],
        mapping=mock_mapping,
    )
    record_model = record_api_models[0]
    assert record_model.fields["prompt"] == "Hello World, how are you?"
    assert record_model.external_id == record_id


def test_ingest_record_from_dict_with_mapped_metadata_vectors(dataset):
    mock_mapping = {
        "test_score": "score",
        "test_vector": "vector",
    }
    record_api_models = dataset.records._ingest_records(
        records=[
            {
                "prompt": "Hello World, how are you?",
                "label": "negative",
                "test_score": 0.9,
                "test_vector": [1, 2, 3],
            }
        ],
        mapping=mock_mapping,
    )

    record = record_api_models[0]
    assert record.fields["prompt"] == "Hello World, how are you?"
    assert record.suggestions[0].value == "negative"
    assert record.vectors[0].vector_values == [1, 2, 3]
    assert record.vectors[0].name == "vector"
    assert record.metadata[0].value == 0.9
    assert record.metadata[0].name == "score"


def test_ingest_record_from_dict_with_mapping_multiple():
    settings = rg.Settings(
        fields=[rg.TextField(name="prompt_field")],
        questions=[
            rg.LabelQuestion(name="label", labels=["negative", "positive"]),
            rg.TextQuestion(name="prompt_question"),
        ],
    )
    workspace = rg.Workspace(name="workspace", id=uuid4())
    dataset = rg.Dataset(
        name="test_dataset",
        settings=settings,
        workspace=workspace,
    )
    mapping = {
        "my_prompt": ("prompt_field", "prompt_question"),
    }
    record_api_models = dataset.records._ingest_records(
        records=[
            {
                "my_prompt": "What is the capital of France?",
                "label": "positive",
            }
        ],
        mapping=mapping,
    )
    record = record_api_models[0]
    suggestions = [s.value for s in record.suggestions]
    assert record.fields["prompt_field"] == "What is the capital of France?"
    assert "positive" in suggestions
    assert "What is the capital of France?" in suggestions



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import os
from functools import cached_property
from pathlib import Path
from typing import List, Optional, TYPE_CHECKING, Dict, Union, Iterator, Sequence
from uuid import UUID

from argilla._exceptions import SettingsError, ArgillaAPIError, ArgillaSerializeError
from argilla._models._dataset import DatasetModel
from argilla._resource import Resource
from argilla.settings._field import TextField
from argilla.settings._metadata import MetadataType, MetadataField
from argilla.settings._question import QuestionType, question_from_model, question_from_dict, QuestionPropertyBase
from argilla.settings._vector import VectorField

if TYPE_CHECKING:
    from argilla.datasets import Dataset


__all__ = ["Settings"]


class Settings(Resource):
    """
    Settings class for Argilla Datasets.

    This class is used to define the representation of a Dataset within the UI.
    """

    def __init__(
        self,
        fields: Optional[List[TextField]] = None,
        questions: Optional[List[QuestionType]] = None,
        vectors: Optional[List[VectorField]] = None,
        metadata: Optional[List[MetadataType]] = None,
        guidelines: Optional[str] = None,
        allow_extra_metadata: bool = False,
        _dataset: Optional["Dataset"] = None,
    ) -> None:
        """
        Args:
            fields (List[TextField]): A list of TextField objects that represent the fields in the Dataset.
            questions (List[Union[LabelQuestion, MultiLabelQuestion, RankingQuestion, TextQuestion, RatingQuestion]]): A list of Question objects that represent the questions in the Dataset.
            vectors (List[VectorField]): A list of VectorField objects that represent the vectors in the Dataset.
            metadata (List[MetadataField]): A list of MetadataField objects that represent the metadata in the Dataset.
            guidelines (str): A string containing the guidelines for the Dataset.
            allow_extra_metadata (bool): A boolean that determines whether or not extra metadata is allowed in the Dataset. Defaults to False.
        """
        super().__init__(client=_dataset._client if _dataset else None)

        self.__questions = questions or []
        self.__fields = SettingsProperties(self, fields)
        self.__vectors = SettingsProperties(self, vectors)
        self.__metadata = SettingsProperties(self, metadata)

        self.__guidelines = self.__process_guidelines(guidelines)
        self.__allow_extra_metadata = allow_extra_metadata

        self._dataset = _dataset

    #####################
    # Properties        #
    #####################

    @property
    def fields(self) -> "SettingsProperties":
        return self.__fields

    @fields.setter
    def fields(self, fields: List[TextField]):
        self.__fields = SettingsProperties(self, fields)

    @property
    def questions(self) -> List[QuestionType]:
        return self.__questions

    @questions.setter
    def questions(self, questions: List[QuestionType]):
        self.__questions = questions

    @property
    def vectors(self) -> "SettingsProperties":
        return self.__vectors

    @vectors.setter
    def vectors(self, vectors: List[VectorField]):
        self.__vectors = SettingsProperties(self, vectors)

    @property
    def metadata(self) -> "SettingsProperties":
        return self.__metadata

    @metadata.setter
    def metadata(self, metadata: List[MetadataType]):
        self.__metadata = SettingsProperties(self, metadata)

    @property
    def guidelines(self) -> str:
        return self.__guidelines

    @guidelines.setter
    def guidelines(self, guidelines: str):
        self.__guidelines = self.__process_guidelines(guidelines)

    @property
    def allow_extra_metadata(self) -> bool:
        return self.__allow_extra_metadata

    @allow_extra_metadata.setter
    def allow_extra_metadata(self, value: bool):
        self.__allow_extra_metadata = value

    @property
    def dataset(self) -> "Dataset":
        return self._dataset

    @dataset.setter
    def dataset(self, dataset: "Dataset"):
        self._dataset = dataset
        self._client = dataset._client

    @cached_property
    def schema(self) -> dict:
        schema_dict = {}

        for field in self.fields:
            schema_dict[field.name] = field

        for question in self.questions:
            schema_dict[question.name] = question

        for vector in self.vectors:
            schema_dict[vector.name] = vector

        for metadata in self.metadata:
            schema_dict[metadata.name] = metadata

        return schema_dict

    @cached_property
    def schema_by_id(self) -> Dict[UUID, Union[TextField, QuestionType, MetadataType, VectorField]]:
        return {v.id: v for v in self.schema.values()}

    def validate(self) -> None:
        self._validate_empty_settings()
        self._validate_duplicate_names()

    #####################
    #  Public methods   #
    #####################

    def get(self) -> "Settings":
        self.fields = self._fetch_fields()
        self.questions = self._fetch_questions()
        self.vectors = self._fetch_vectors()
        self.metadata = self._fetch_metadata()
        self.__get_dataset_related_attributes()

        self._update_last_api_call()
        return self

    def create(self) -> "Settings":
        self.validate()

        self._update_dataset_related_attributes()
        self.__fields.create()
        self._create_questions()
        self.__vectors.create()
        self.__metadata.create()

        self._update_last_api_call()
        return self

    def update(self) -> "Resource":
        self.validate()

        self._update_dataset_related_attributes()
        self.__fields.update()
        self.__vectors.update()
        self.__metadata.update()
        # self.questions.update()

        self._update_last_api_call()
        return self

    def question_by_name(self, question_name: str) -> QuestionType:
        for question in self.questions:
            if question.name == question_name:
                return question
        raise ValueError(f"Question with name {question_name} not found")

    def question_by_id(self, question_id: UUID) -> QuestionType:
        property = self.schema_by_id.get(question_id)
        if isinstance(property, QuestionPropertyBase):
            return property
        raise ValueError(f"Question with id {question_id} not found")

    def serialize(self):
        try:
            return {
                "guidelines": self.guidelines,
                "questions": self.__serialize_questions(self.questions),
                "fields": self.__fields.serialize(),
                "vectors": self.vectors.serialize(),
                "metadata": self.metadata.serialize(),
                "allow_extra_metadata": self.allow_extra_metadata,
            }
        except Exception as e:
            raise ArgillaSerializeError(f"Failed to serialize the settings. {e.__class__.__name__}") from e

    def to_json(self, path: Union[Path, str]) -> None:
        """Save the settings to a file on disk

        Parameters:
            path (str): The path to save the settings to
        """
        if not isinstance(path, Path):
            path = Path(path)
        if path.exists():
            raise FileExistsError(f"File {path} already exists")
        with open(path, "w") as file:
            json.dump(self.serialize(), file)

    @classmethod
    def from_json(cls, path: Union[Path, str]) -> "Settings":
        """Load the settings from a file on disk"""

        with open(path, "r") as file:
            settings_dict = json.load(file)

        fields = settings_dict.get("fields", [])
        vectors = settings_dict.get("vectors", [])
        metadata = settings_dict.get("metadata", [])
        guidelines = settings_dict.get("guidelines")
        allow_extra_metadata = settings_dict.get("allow_extra_metadata")

        questions = [question_from_dict(question) for question in settings_dict.get("questions", [])]
        fields = [TextField.from_dict(field) for field in fields]
        vectors = [VectorField.from_dict(vector) for vector in vectors]
        metadata = [MetadataField.from_dict(metadata) for metadata in metadata]

        return cls(
            questions=questions,
            fields=fields,
            vectors=vectors,
            metadata=metadata,
            guidelines=guidelines,
            allow_extra_metadata=allow_extra_metadata,
        )

    def __eq__(self, other: "Settings") -> bool:
        return self.serialize() == other.serialize()  # TODO: Create proper __eq__ methods for fields and questions

    #####################
    #  Repr Methods     #
    #####################

    def __repr__(self) -> str:
        return (
            f"Settings(guidelines={self.guidelines}, allow_extra_metadata={self.allow_extra_metadata}, "
            f"fields={self.fields}, questions={self.questions}, vectors={self.vectors}, metadata={self.metadata})"
        )

    #####################
    #  Private methods  #
    #####################

    def _fetch_fields(self) -> List[TextField]:
        models = self._client.api.fields.list(dataset_id=self._dataset.id)
        return [TextField.from_model(model) for model in models]

    def _fetch_questions(self) -> List[QuestionType]:
        models = self._client.api.questions.list(dataset_id=self._dataset.id)
        return [question_from_model(model) for model in models]

    def _fetch_vectors(self) -> List[VectorField]:
        models = self.dataset._client.api.vectors.list(self.dataset.id)
        return [VectorField.from_model(model) for model in models]

    def _fetch_metadata(self) -> List[MetadataType]:
        models = self._client.api.metadata.list(dataset_id=self._dataset.id)
        return [MetadataField.from_model(model) for model in models]

    def __get_dataset_related_attributes(self):
        # This flow may be a bit weird, but it's the only way to update the dataset related attributes
        # Everything is point that we should have several settings-related endpoints in the API to handle this.
        # POST /api/v1/datasets/{dataset_id}/settings
        # {
        #   "guidelines": ....,
        #   "allow_extra_metadata": ....,
        # }
        # But this is not implemented yet, so we need to update the dataset model directly
        dataset_model = self._client.api.datasets.get(self._dataset.id)

        self.guidelines = dataset_model.guidelines
        self.allow_extra_metadata = dataset_model.allow_extra_metadata

    def _update_dataset_related_attributes(self):
        # This flow may be a bit weird, but it's the only way to update the dataset related attributes
        # Everything is point that we should have several settings-related endpoints in the API to handle this.
        # POST /api/v1/datasets/{dataset_id}/settings
        # {
        #   "guidelines": ....,
        #   "allow_extra_metadata": ....,
        # }
        # But this is not implemented yet, so we need to update the dataset model directly
        dataset_model = DatasetModel(
            id=self._dataset.id,
            name=self._dataset.name,
            guidelines=self.guidelines,
            allow_extra_metadata=self.allow_extra_metadata,
        )
        self._client.api.datasets.update(dataset_model)

    def _create_questions(self) -> None:
        for question in self.__questions:
            try:
                question_model = self._client.api.questions.create(
                    dataset_id=self._dataset.id, question=question._model
                )
                question._model = question_model
            except ArgillaAPIError as e:
                raise SettingsError(f"Failed to create question {question.name}") from e

    def _validate_empty_settings(self):
        if not all([self.fields, self.questions]):
            message = "Fields and questions are required"
            raise SettingsError(message=message)

    def _validate_duplicate_names(self) -> None:
        dataset_properties_by_name = {}

        for properties in [self.fields, self.questions, self.vectors, self.metadata]:
            for property in properties:
                if property.name in dataset_properties_by_name:
                    raise SettingsError(
                        f"names of dataset settings must be unique, "
                        f"but the name {property.name!r} is used by {type(property).__name__!r} and {type(dataset_properties_by_name[property.name]).__name__!r} "
                    )
                dataset_properties_by_name[property.name] = property

    def __process_guidelines(self, guidelines):
        if guidelines is None:
            return guidelines

        if not isinstance(guidelines, str):
            raise SettingsError("Guidelines must be a string or a path to a file")

        if os.path.exists(guidelines):
            with open(guidelines, "r") as file:
                return file.read()

        return guidelines

    def __serialize_questions(self, questions: List[QuestionType]):
        return [question.serialize() for question in questions]


Property = Union[TextField, VectorField, MetadataType, QuestionType]


class SettingsProperties(Sequence[Property]):
    """A collection of properties (fields, questions, vectors and metadata) for a dataset settings object.

    This class is used to store the properties of a dataset settings object
    """

    def __init__(self, settings: "Settings", properties: List[Property]):
        self._properties_by_name = {}
        self._settings = settings

        for property in properties or []:
            self.add(property)

    def __getitem__(self, key: Union[str, int]) -> Optional[Property]:
        if isinstance(key, int):
            return list(self._properties_by_name.values())[key]
        return self._properties_by_name.get(key)

    def __iter__(self) -> Iterator[Property]:
        return iter(self._properties_by_name.values())

    def __len__(self):
        return len(self._properties_by_name)

    def __eq__(self, other):
        """Check if two instances are equal. Overloads the == operator."""
        if not isinstance(other, SettingsProperties):
            return False
        return self._properties_by_name == other._properties_by_name

    def add(self, property: Property) -> Property:
        self._validate_new_property(property)
        self._properties_by_name[property.name] = property
        setattr(self, property.name, property)
        return property

    def create(self):
        for property in self:
            try:
                property.dataset = self._settings.dataset
                property.create()
            except ArgillaAPIError as e:
                raise SettingsError(f"Failed to create property {property.name!r}: {e.message}") from e

    def update(self):
        for item in self:
            try:
                item.dataset = self._settings.dataset
                item.update() if item.id else item.create()
            except ArgillaAPIError as e:
                raise SettingsError(f"Failed to update {item.name!r}: {e.message}") from e

    def serialize(self) -> List[dict]:
        return [property.serialize() for property in self]

    def _validate_new_property(self, property: Property) -> None:
        if property.name in self._properties_by_name:
            raise ValueError(f"Property with name {property.name!r} already exists in the collection")

        if property.name in dir(self):
            raise ValueError(f"Property with name {property.name!r} conflicts with an existing attribute")



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from uuid import uuid4

from datasets import Value, Sequence

import argilla as rg
from argilla.records._io import HFDatasetsIO


class TestHFDatasetsIO:
    def test_to_datasets_with_partial_values_in_records(self):
        records = [
            rg.Record(fields={"field": "The field"}, metadata={"a": "a"}),
            rg.Record(fields={"field": "Other field", "other": "Field"}, metadata={"b": "b"}),
            rg.Record(fields={"field": "Again"}, suggestions=[rg.Suggestion("question", value="value")]),
            rg.Record(
                fields={"field": "Field"}, responses=[rg.Response("other_question", value="value", user_id=uuid4())]
            ),
            rg.Record(
                fields={"field": "The record field including more type of responses"},
                suggestions=[
                    rg.Suggestion("rating", value=1),
                    rg.Suggestion("ranking", value=["value1", "value2"]),
                    rg.Suggestion("spans", value=[{"start": 0, "end": 10, "label": "test"}]),
                ],
                responses=[
                    rg.Response("rating", value=1, user_id=uuid4()),
                    rg.Response("ranking", value=["value1", "value2"], user_id=uuid4()),
                    rg.Response("spans", value=[{"start": 0, "end": 10, "label": "test"}], user_id=uuid4()),
                ],
            ),
        ]

        ds = HFDatasetsIO.to_datasets(records)
        assert ds.features == {
            "_server_id": Value(dtype="null", id=None),
            "a": Value(dtype="string", id=None),
            "b": Value(dtype="string", id=None),
            "field": Value(dtype="string", id=None),
            "id": Value(dtype="string", id=None),
            "other": Value(dtype="string", id=None),
            "other_question.responses": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
            "other_question.responses.users": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
            "question.suggestion": Value(dtype="string", id=None),
            "question.suggestion.agent": Value(dtype="null", id=None),
            "question.suggestion.score": Value(dtype="null", id=None),
            "ranking.responses": Sequence(
                feature=Sequence(feature=Value(dtype="string", id=None), length=-1, id=None), length=-1, id=None
            ),
            "ranking.responses.users": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
            "ranking.suggestion": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
            "ranking.suggestion.agent": Value(dtype="null", id=None),
            "ranking.suggestion.score": Value(dtype="null", id=None),
            "rating.responses": Sequence(feature=Value(dtype="int64", id=None), length=-1, id=None),
            "rating.responses.users": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
            "rating.suggestion": Value(dtype="int64", id=None),
            "rating.suggestion.agent": Value(dtype="null", id=None),
            "rating.suggestion.score": Value(dtype="null", id=None),
            "spans.responses": [
                [
                    {
                        "end": Value(dtype="int64", id=None),
                        "label": Value(dtype="string", id=None),
                        "start": Value(dtype="int64", id=None),
                    }
                ]
            ],
            "spans.responses.users": Sequence(feature=Value(dtype="string", id=None), length=-1, id=None),
            "spans.suggestion": [
                {
                    "end": Value(dtype="int64", id=None),
                    "label": Value(dtype="string", id=None),
                    "start": Value(dtype="int64", id=None),
                }
            ],
            "spans.suggestion.agent": Value(dtype="null", id=None),
            "spans.suggestion.score": Value(dtype="null", id=None),
        }



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, TYPE_CHECKING

from argilla._api._vectors import VectorsAPI
from argilla._models import VectorFieldModel
from argilla._resource import Resource
from argilla.client import Argilla

if TYPE_CHECKING:
    from argilla import Dataset

__all__ = ["VectorField"]


class VectorField(Resource):
    """Vector field for use in Argilla `Dataset` `Settings`"""

    _model: VectorFieldModel
    _api: VectorsAPI
    _dataset: "Dataset"

    def __init__(
        self,
        name: str,
        dimensions: int,
        title: Optional[str] = None,
        _client: Optional["Argilla"] = None,
    ) -> None:
        """Vector field for use in Argilla `Dataset` `Settings`

        Parameters:
            name (str): The name of the field
            dimensions (int): The number of dimensions in the vector
            title (Optional[str], optional): The title of the field. Defaults to None.
        """
        client = _client or Argilla._get_default()
        super().__init__(api=client.api.vectors, client=client)
        self._model = VectorFieldModel(name=name, title=title, dimensions=dimensions)
        self._dataset = None

    @property
    def name(self) -> str:
        return self._model.name

    @name.setter
    def name(self, value: str) -> None:
        self._model.name = value

    @property
    def title(self) -> Optional[str]:
        return self._model.title

    @title.setter
    def title(self, value: Optional[str]) -> None:
        self._model.title = value

    @property
    def dimensions(self) -> int:
        return self._model.dimensions

    @dimensions.setter
    def dimensions(self, value: int) -> None:
        self._model.dimensions = value

    @property
    def dataset(self) -> "Dataset":
        return self._dataset

    @dataset.setter
    def dataset(self, value: "Dataset") -> None:
        self._dataset = value
        self._model.dataset_id = self._dataset.id

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(name={self.name}, title={self.title}, dimensions={self.dimensions})"

    @classmethod
    def from_model(cls, model: VectorFieldModel) -> "VectorField":
        instance = cls(name=model.name, dimensions=model.dimensions)
        instance._model = model

        return instance

    @classmethod
    def from_dict(cls, data: dict) -> "VectorField":
        model = VectorFieldModel(**data)
        return cls.from_model(model=model)



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid
import pytest

import argilla as rg


@pytest.fixture
def dataset(client: rg.Argilla) -> rg.Dataset:
    workspace = client.workspaces[0]
    mock_dataset_name = f"test_delete_records_{uuid.uuid1()}"
    settings = rg.Settings(
        allow_extra_metadata=True,
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.TextQuestion(name="label", use_markdown=False),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        workspace=workspace.name,
        settings=settings,
        client=client,
    )
    dataset.create()
    return dataset


def test_delete_records(client: rg.Argilla, dataset: rg.Dataset):
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
        },
    ]

    dataset.records.log(records=mock_data)
    records_to_delete = list(dataset.records)[:2]
    dataset.records.delete(records_to_delete)
    dataset_records = list(dataset.records)

    assert len(dataset_records) == 1
    assert dataset_records[0].id == str(mock_data[2]["id"])

    for record in dataset_records:
        assert record.id not in [record.id for record in records_to_delete]


def test_delete_single_record(client: rg.Argilla, dataset: rg.Dataset):
    mock_data = [
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
        },
        {
            "text": "Hello World, how are you?",
            "label": "negative",
            "id": uuid.uuid4(),
        },
    ]

    dataset.records.log(records=mock_data)
    records_to_delete = [list(dataset.records)[1]]
    dataset.records.delete(records_to_delete)
    dataset_records = list(dataset.records)

    assert len(dataset_records) == 2
    assert dataset_records[0].id == str(mock_data[0]["id"])
    assert dataset_records[1].id == str(mock_data[2]["id"])
    assert mock_data[1]["id"] not in [record.id for record in dataset_records]



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from argilla.records import Filter


class TestFilters:
    def test_filter_by_responses_status(self):
        test_filter = Filter(("response.status", "in", ["submitted", "discard"]))
        assert test_filter.api_model().model_dump(by_alias=True) == {
            "type": "and",
            "and": [
                {
                    "scope": {"entity": "response", "property": "status", "question": None},
                    "type": "terms",
                    "values": ["submitted", "discard"],
                }
            ],
        }



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Optional, Union, List, TYPE_CHECKING

from argilla._api._metadata import MetadataAPI
from argilla._exceptions import MetadataError
from argilla._models import (
    MetadataPropertyType,
    TermsMetadataPropertySettings,
    FloatMetadataPropertySettings,
    IntegerMetadataPropertySettings,
    MetadataFieldModel,
)
from argilla._resource import Resource
from argilla.client import Argilla

if TYPE_CHECKING:
    from argilla import Dataset

__all__ = [
    "TermsMetadataProperty",
    "FloatMetadataProperty",
    "IntegerMetadataProperty",
    "MetadataType",
]


class MetadataPropertyBase(Resource):
    _model: MetadataFieldModel
    _api: MetadataAPI

    _dataset: "Dataset"

    def __init__(self, client: Optional[Argilla] = None) -> None:
        client = client or Argilla._get_default()
        super().__init__(client=client, api=client.api.metadata)

    @property
    def name(self) -> str:
        return self._model.name

    @name.setter
    def name(self, value: str) -> None:
        self._model.name = value

    @property
    def title(self) -> Optional[str]:
        return self._model.title

    @title.setter
    def title(self, value: Optional[str]) -> None:
        self._model.title = value

    @property
    def visible_for_annotators(self) -> Optional[bool]:
        return self._model.visible_for_annotators

    @visible_for_annotators.setter
    def visible_for_annotators(self, value: Optional[bool]) -> None:
        self._model.visible_for_annotators = value

    @property
    def dataset(self) -> Optional["Dataset"]:
        return self._dataset

    @dataset.setter
    def dataset(self, value: "Dataset") -> None:
        self._dataset = value
        self._model.dataset_id = value.id

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(name={self.name}, title={self.title}, dimensions={self.visible_for_annotators})"
        )


class TermsMetadataProperty(MetadataPropertyBase):
    def __init__(
        self,
        name: str,
        options: Optional[List[str]] = None,
        title: Optional[str] = None,
        visible_for_annotators: Optional[bool] = True,
        client: Optional[Argilla] = None,
    ) -> None:
        """Create a metadata field with terms settings.

        Parameters:
            name (str): The name of the metadata field
            options (Optional[List[str]]): The list of options
            title (Optional[str]): The title of the metadata field
            visible_for_annotators (Optional[bool]): Whether the metadata field is visible for annotators
        Raises:
            MetadataError: If an error occurs while defining metadata settings
        """
        super().__init__(client=client)

        try:
            settings = TermsMetadataPropertySettings(values=options, type=MetadataPropertyType.terms)
        except ValueError as e:
            raise MetadataError(f"Error defining metadata settings for {name}") from e

        self._model = MetadataFieldModel(
            name=name,
            type=MetadataPropertyType.terms,
            title=title,
            settings=settings,
            visible_for_annotators=visible_for_annotators,
        )

    @property
    def options(self) -> Optional[List[str]]:
        return self._model.settings.values

    @options.setter
    def options(self, value: list[str]) -> None:
        self._model.settings.values = value

    @classmethod
    def from_model(cls, model: MetadataFieldModel) -> "TermsMetadataProperty":
        instance = TermsMetadataProperty(name=model.name)
        instance._model = model

        return instance


class FloatMetadataProperty(MetadataPropertyBase):
    def __init__(
        self,
        name: str,
        min: Optional[float] = None,
        max: Optional[float] = None,
        title: Optional[str] = None,
        client: Optional[Argilla] = None,
    ) -> None:
        """Create a metadata field with float settings.

        Parameters:
            name (str): The name of the metadata field
            min (Optional[float]): The minimum value
            max (Optional[float]): The maximum value
            title (Optional[str]): The title of the metadata field
            client (Optional[Argilla]): The client to use for API requests
        Raises:
            MetadataError: If an error occurs while defining metadata settings

        """

        super().__init__(client=client)

        try:
            settings = FloatMetadataPropertySettings(min=min, max=max, type=MetadataPropertyType.float)
        except ValueError as e:
            raise MetadataError(f"Error defining metadata settings for {name}") from e

        self._model = MetadataFieldModel(
            name=name,
            type=MetadataPropertyType.float,
            title=title,
            settings=settings,
        )

    @property
    def min(self) -> Optional[int]:
        return self._model.settings.min

    @min.setter
    def min(self, value: Optional[int]) -> None:
        self._model.settings.min = value

    @property
    def max(self) -> Optional[int]:
        return self._model.settings.max

    @max.setter
    def max(self, value: Optional[int]) -> None:
        self._model.settings.max = value

    @classmethod
    def from_model(cls, model: MetadataFieldModel) -> "FloatMetadataProperty":
        instance = FloatMetadataProperty(name=model.name)
        instance._model = model

        return instance


class IntegerMetadataProperty(MetadataPropertyBase):
    def __init__(
        self,
        name: str,
        min: Optional[int] = None,
        max: Optional[int] = None,
        title: Optional[str] = None,
        client: Optional[Argilla] = None,
    ) -> None:
        """Create a metadata field with integer settings.

        Parameters:
            name (str): The name of the metadata field
            min (Optional[int]): The minimum value
            max (Optional[int]): The maximum value
            title (Optional[str]): The title of the metadata field
        Raises:
            MetadataError: If an error occurs while defining metadata settings
        """
        super().__init__(client=client)

        try:
            settings = IntegerMetadataPropertySettings(min=min, max=max, type=MetadataPropertyType.integer)
        except ValueError as e:
            raise MetadataError(f"Error defining metadata settings for {name}") from e

        self._model = MetadataFieldModel(
            name=name,
            type=MetadataPropertyType.integer,
            title=title,
            settings=settings,
        )

    @property
    def min(self) -> Optional[int]:
        return self._model.settings.min

    @min.setter
    def min(self, value: Optional[int]) -> None:
        self._model.settings.min = value

    @property
    def max(self) -> Optional[int]:
        return self._model.settings.max

    @max.setter
    def max(self, value: Optional[int]) -> None:
        self._model.settings.max = value

    @classmethod
    def from_model(cls, model: MetadataFieldModel) -> "IntegerMetadataProperty":
        instance = IntegerMetadataProperty(name=model.name)
        instance._model = model

        return instance


MetadataType = Union[
    TermsMetadataProperty,
    FloatMetadataProperty,
    IntegerMetadataProperty,
]


class MetadataField:
    @classmethod
    def from_model(cls, model: MetadataFieldModel) -> MetadataType:
        switch = {
            MetadataPropertyType.terms: TermsMetadataProperty,
            MetadataPropertyType.float: FloatMetadataProperty,
            MetadataPropertyType.integer: IntegerMetadataProperty,
        }
        metadata_type = model.type
        try:
            return switch[metadata_type].from_model(model)
        except KeyError as e:
            raise MetadataError(f"Unknown metadata property type: {metadata_type}") from e

    @classmethod
    def from_dict(cls, data: dict) -> MetadataType:
        switch = {
            MetadataPropertyType.terms: TermsMetadataProperty,
            MetadataPropertyType.float: FloatMetadataProperty,
            MetadataPropertyType.integer: IntegerMetadataProperty,
        }
        metadata_type = data["type"]
        try:
            metadata_model = MetadataFieldModel(**data)
            return switch[metadata_type].from_model(metadata_model)
        except KeyError as e:
            raise MetadataError(f"Unknown metadata property type: {metadata_type}") from e



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import random
import uuid
from string import ascii_lowercase

import pytest

import argilla as rg
from argilla import Record
from argilla._models import RecordModel


@pytest.fixture
def dataset(client: rg.Argilla) -> rg.Dataset:
    workspace = client.workspaces[0]
    mock_dataset_name = "".join(random.choices(ascii_lowercase, k=16))
    settings = rg.Settings(
        allow_extra_metadata=True,
        fields=[
            rg.TextField(name="text"),
        ],
        questions=[
            rg.TextQuestion(name="label", use_markdown=False),
        ],
    )
    dataset = rg.Dataset(
        name=mock_dataset_name,
        workspace=workspace.name,
        settings=settings,
        client=client,
    )
    dataset.create()
    return dataset


class TestUpdateSuggestions:
    def test_update_records_suggestions_from_data(self, client: rg.Argilla, dataset: rg.Dataset):
        mock_data = [
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
        ]
        updated_mock_data = [
            {
                "text": r["text"],
                "label": "positive",
                "id": r["id"],
            }
            for r in mock_data
        ]

        dataset.records.log(records=mock_data)
        dataset.records.log(records=updated_mock_data)
        dataset_records = list(dataset.records)

        assert dataset_records[0].id == str(mock_data[0]["id"])
        assert dataset_records[1].id == str(mock_data[1]["id"])
        assert dataset_records[2].id == str(mock_data[2]["id"])
        for record in dataset.records(with_suggestions=True):
            assert record.suggestions["label"].value == "positive"

    @pytest.mark.skip(reason="This test is failing because the backend expects the fields to be present in the data.")
    def test_update_records_without_fields(self, client: rg.Argilla, dataset: rg.Dataset):
        mock_data = [
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
        ]

        updated_mock_data = mock_data.copy()
        for record in updated_mock_data:
            record.pop("text")
            record["label"] = "positive"
        dataset.records.log(records=mock_data)
        dataset.records.log(records=updated_mock_data)

        for i, record in enumerate(dataset.records(with_suggestions=True)):
            assert record.suggestions["label"].value == updated_mock_data[i]["label"]

    def test_update_records_add_suggestions(self, client: rg.Argilla, dataset: rg.Dataset):
        mock_data = [
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
        ]
        dataset.records.log(records=mock_data)

        updated_records = []

        for record in dataset.records(with_suggestions=True):
            record.suggestions.add(
                rg.Suggestion(
                    question_name="label",
                    value="positive",
                )
            )
            updated_records.append(record)

        dataset.records.log(records=updated_records)

        for record in dataset.records(with_suggestions=True):
            assert record.suggestions["label"].value == "positive"


class TestUpdateResponses:
    def test_update_records_add_responses(self, client: rg.Argilla, dataset: rg.Dataset):
        mock_data = [
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
            {
                "text": "Hello World, how are you?",
                "label": "negative",
                "id": uuid.uuid4(),
            },
        ]
        dataset.records.log(records=mock_data)

        updated_records = []

        for record in dataset.records(with_suggestions=True):
            record.responses.add(
                rg.Response(
                    question_name="label",
                    value="positive",
                    user_id=client.users[0].id,
                )
            )
            updated_records.append(record)

        dataset.records.log(records=updated_records)

        for record in dataset.records(with_suggestions=True):
            assert record.responses["label"][-1].value == "positive"



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Dict, List, Optional

from pydantic import BaseModel, Field


class ParameterType(str, Enum):
    """Parameter types are the different 'sub' values of a records attribute.
    For example, the value, score, or agent of a suggestion."""

    VALUE = "value"
    SCORE = "score"
    AGENT = "agent"

    @classmethod
    def values(cls) -> List[str]:
        return [param.value for param in cls]


class AttributeType(str, Enum):
    """Attribute types are the different types of attributes a record can have."""

    FIELD = "field"
    SUGGESTION = "suggestion"
    RESPONSE = "response"
    METADATA = "metadata"
    VECTOR = "vector"
    ID = "id"

    @classmethod
    def values(cls) -> List[str]:
        return [attr.value for attr in cls]


class AttributeParameter(BaseModel):
    """Attribute parameters are reference connections between the `ParameterType`'s of a records `AttributeType`.
    The connect the source key to the sub value of the attribute. i.e. column name 'score' to the 'score' of a suggestion.
    """

    parameter_type: ParameterType = ParameterType.VALUE
    source: str


class AttributeRoute(BaseModel):
    """AttributeRoute is a reference connection between a record's attribute and a source value in the data.
    It connects the source key with the attribute name and type. For example, connecting the columns 'score'
    and 'y' to the 'score' and 'value' of a suggestion.
    """

    source: str
    name: str
    type: Optional[AttributeType] = None
    parameters: List[AttributeParameter] = []

    def set_parameter(self, parameter: AttributeParameter):
        """Set a parameter for the route.
        An existing parameter with same parameter type will be replaced by this new one.
        """
        for p in self.parameters:
            if p.parameter_type == parameter.parameter_type:
                self.parameters.remove(p)
                break
        self.parameters.append(parameter)


class RecordAttributesMap(BaseModel):
    """RecordAttributesMap is a representation of a record attribute mapping that is used to parse data into a record."""

    suggestion: Dict[str, AttributeRoute] = Field(default_factory=dict)
    response: Dict[str, AttributeRoute] = Field(default_factory=dict)
    field: Dict[str, AttributeRoute] = Field(default_factory=dict)
    metadata: Dict[str, AttributeRoute] = Field(default_factory=dict)
    vector: Dict[str, AttributeRoute] = Field(default_factory=dict)

    id: AttributeRoute = AttributeRoute(source="id", name="id", type=AttributeType.ID)

    def _get_routes_group_by_type(self, type: AttributeType):
        """Utility method to facilitate getting the routes by type."""
        return {
            AttributeType.SUGGESTION: self.suggestion,
            AttributeType.RESPONSE: self.response,
            AttributeType.FIELD: self.field,
            AttributeType.METADATA: self.metadata,
            AttributeType.VECTOR: self.vector,
            AttributeType.ID: self.id,
        }[type]

    def get_by_name_and_type(self, name: str, type: AttributeType) -> Optional[AttributeRoute]:
        """Utility method to get a route by name and type"""
        if name == "id" and AttributeType.ID:
            return self.id
        return self._get_routes_group_by_type(type).get(name)

    def add_route(self, attribute_route: AttributeRoute) -> None:
        """Utility method to get a new mapping route"""
        if attribute_route.type == AttributeType.ID:
            self.id = attribute_route
        else:
            self._get_routes_group_by_type(attribute_route.type)[attribute_route.name] = attribute_route



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Sequence, Union, Tuple
from uuid import UUID

from argilla._exceptions import RecordsIngestionError
from argilla.records._resource import Record
from argilla.responses import Response
from argilla.settings import TextField, VectorField
from argilla.settings._metadata import MetadataPropertyBase
from argilla.settings._question import QuestionPropertyBase
from argilla.suggestions import Suggestion
from argilla.records._mapping._routes import (
    AttributeRoute,
    RecordAttributesMap,
    AttributeType,
    ParameterType,
    AttributeParameter,
)

if TYPE_CHECKING:
    from argilla.datasets import Dataset


class IngestedRecordMapper:
    """IngestedRecordMapper is a class that is used to map data into a record object.
    It maps values in ingested data to the appropriate record attributes, based on the user
    provided mapping and the schema of the dataset.

    The Mapper builds and uses a `RecordAttributesMap` object to map the data to the appropriate record attributes.

    Attributes:
        dataset: The dataset the record will be added to.
        mapping: A dictionary mapping from source data keys/ columns to Argilla fields, questions, ids, etc.
        user_id: The user id to associate with the record responses.
    """

    mapping: RecordAttributesMap = None

    def __init__(
        self,
        dataset: "Dataset",
        user_id: UUID,
        mapping: Optional[Dict[str, Union[str, Sequence[str]]]] = None,
    ):
        self._dataset = dataset
        self._schema = dataset.schema
        self.user_id = user_id

        mapping = mapping or {}
        default_mapping = self._schematize_default_attributes()
        self.mapping = self._schematize_mapped_attributes(mapping=mapping, default_mapping=default_mapping)

    def __call__(self, data: Dict[str, Any], user_id: Optional[UUID] = None) -> Record:
        """Maps a dictionary of data to a record object.

        Parameters:
            data: A dictionary representing the record.
            user_id: The user id to associate with the record responses.

        Returns:
            Record: The record object.

        """

        record_id = data.get(self.mapping.id.source)
        suggestions = self._map_suggestions(data=data, mapping=self.mapping.suggestion)
        responses = self._map_responses(data=data, user_id=user_id or self.user_id, mapping=self.mapping.response)
        fields = self._map_attributes(data=data, mapping=self.mapping.field)
        metadata = self._map_attributes(data=data, mapping=self.mapping.metadata)
        vectors = self._map_attributes(data=data, mapping=self.mapping.vector)

        return Record(
            id=record_id,
            fields=fields,
            vectors=vectors,
            metadata=metadata,
            suggestions=suggestions,
            responses=responses,
            _dataset=self._dataset,
        )

    ##########################################
    # Private helper functions - Build Mapping
    ##########################################

    def _schematize_mapped_attributes(
        self,
        mapping: Dict[str, Union[str, Sequence[str]]],
        default_mapping: RecordAttributesMap,
    ) -> RecordAttributesMap:
        """Extends the default mapping with a schematized mapping object provided from a dict"""

        for source_key, value in mapping.items():
            mapped_attributes = [value] if isinstance(value, str) else list(value)
            for attr_mapping in mapped_attributes:
                attr_name, attr_type, parameter = self._parse_dot_notation(attr_mapping)

                attr_type = AttributeType(attr_type or AttributeType.SUGGESTION)
                parameter = AttributeParameter(parameter_type=parameter or ParameterType.VALUE, source=source_key)

                attr_route = default_mapping.get_by_name_and_type(name=attr_name, type=attr_type)
                if attr_route:
                    attr_route.source = source_key
                    attr_route.set_parameter(parameter)
                else:
                    attr_route = AttributeRoute(
                        name=attr_name,
                        source=source_key,
                        type=attr_type,
                        parameters=[parameter],
                    )
                    attr_route = self._select_attribute_type(attribute_route=attr_route)
                    default_mapping.add_route(attr_route)

        return default_mapping

    def _parse_dot_notation(self, attribute_mapping: str) -> Tuple[str, Optional[str], Optional[str]]:
        """Parses a string in the format of 'attribute.type.parameter' into its attribute parts parts using regex."""

        available_attributes = list(self._schema.keys()) + ["id"]
        available_parameters = ParameterType.values()
        available_types = AttributeType.values()

        # The pattern is in the format of 'attribute[.type[.parameter]]' where type and parameter are optional.
        pattern = re.compile(
            rf"^({'|'.join(available_attributes)})"
            rf"(?:\.({'|'.join(available_types)}))?"
            rf"(?:\.({'|'.join(available_parameters)}))?$"
        )

        match = pattern.match(attribute_mapping)
        if not match:
            raise ValueError(
                f"Invalid attribute mapping format: {attribute_mapping}. "
                "Attribute mapping must be in the format of 'attribute[.type[.parameter]]'."
                f"Available attributes: {available_attributes}, types: {available_types}, parameters: {available_parameters}."
            )
        attr_name, type_, parameter = match.groups()
        return attr_name, type_, parameter

    def _select_attribute_type(self, attribute_route: AttributeRoute) -> AttributeRoute:
        """Selects the attribute type based on the schema item and the attribute type.
        This method implements the logic to infer the attribute type based on the schema item if the attribute type is not provided.
        If the attribute type is not provided, it will be inferred based on the schema item.
        """
        schema_item = self._schema.get(attribute_route.name)
        if isinstance(schema_item, QuestionPropertyBase) and (
            attribute_route.type is None or attribute_route.type == AttributeType.SUGGESTION
        ):
            # Suggestions are the default destination for questions.
            attribute_route.type = AttributeType.SUGGESTION
        elif isinstance(schema_item, QuestionPropertyBase) and attribute_route.type == AttributeType.RESPONSE:
            attribute_route.type = AttributeType.RESPONSE
        elif isinstance(schema_item, TextField):
            attribute_route.type = AttributeType.FIELD
        elif isinstance(schema_item, VectorField):
            attribute_route.type = AttributeType.VECTOR
        elif isinstance(schema_item, MetadataPropertyBase):
            attribute_route.type = AttributeType.METADATA
        elif attribute_route.name == "id":
            attribute_route.type = AttributeType.ID
        else:
            raise RecordsIngestionError(f"Mapped attribute is not a valid dataset attribute: {attribute_route.name}.")
        return attribute_route

    def _schematize_default_attributes(self) -> RecordAttributesMap:
        """Creates the mapping with default attribute routes. Uses the schema of the dataset to determine
         the default attributes and add them to the mapping with their names as keys. This means that
         keys in the data that match the names of dataset attributes will be mapped to them by default.

        Returns:
            RecordAttributesMap: The mapping object.
        """
        mapping = RecordAttributesMap()

        for question in self._dataset.settings.questions:
            mapping.suggestion[question.name] = AttributeRoute(
                source=question.name,
                name=question.name,
                type=AttributeType.SUGGESTION,
                parameters=[AttributeParameter(source=question.name)],
            )

        for field in self._dataset.settings.fields:
            mapping.field[field.name] = AttributeRoute(
                source=field.name,
                name=field.name,
                type=AttributeType.FIELD,
            )

        for metadata in self._dataset.settings.metadata:
            mapping.metadata[metadata.name] = AttributeRoute(
                source=metadata.name,
                name=metadata.name,
                type=AttributeType.METADATA,
            )

        for vector in self._dataset.settings.vectors:
            mapping.vector[vector.name] = AttributeRoute(
                source=vector.name,
                name=vector.name,
                type=AttributeType.VECTOR,
            )

        return mapping

    ##########################################
    # Private helper functions - Parse Records
    ##########################################

    def _map_suggestions(self, data: Dict[str, Any], mapping) -> List[Suggestion]:
        """Converts an arbitrary dictionary to a list of Suggestion objects for use by the add or update methods.
        Suggestions can be defined accross multiple columns in the data, so we need to map them to the appropriately.add()

        Parameters:
            data: A dictionary representing the vector.
            mapping: A dictionary mapping from source data keys/ columns to Argilla fields, questions, ids, etc.

        Returns:
             A list of Suggestion objects.

        """
        suggestions = []

        for name, route in mapping.items():
            if route.source not in data:
                continue
            parameters = {param.parameter_type: data.get(param.source) for param in route.parameters}
            if parameters.get(ParameterType.VALUE) is None:
                continue
            schema_item = self._dataset.schema.get(name)
            suggestion = Suggestion(
                **parameters,
                question_name=route.name,
                question_id=schema_item.id,
            )
            suggestions.append(suggestion)

        return suggestions

    def _map_responses(self, data: Dict[str, Any], user_id: UUID, mapping) -> List[Response]:
        """Converts an arbitrary dictionary to a list of Response objects for use by the add or update methods.

        Parameters:
             data: A dictionary representing the vector.
             mapping: A dictionary mapping from source data keys/ columns to Argilla fields, questions, ids, etc.
             user_id: The user id to associate with the record responses.

        Returns:
             A list of Response objects.
        """
        responses = []

        for name, route in mapping.items():
            value = data.get(route.source)
            if value is None:
                continue
            response = Response(
                value=value,
                question_name=name,
                user_id=user_id,
            )
            responses.append(response)

        return responses

    def _map_attributes(self, data: Dict[str, Any], mapping: Dict[str, AttributeRoute]) -> Dict[str, Any]:
        """Converts a dictionary to a dictionary of attributes for use by the add or update methods."""
        attributes = {}

        for name, route in mapping.items():
            if route.source not in data:
                continue
            value = data.get(route.source)
            if value is None:
                continue
            attributes[name] = value

        return attributes



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, Iterable, List, Optional, Sequence, Union
from uuid import UUID

from tqdm import tqdm

from argilla._api import RecordsAPI
from argilla._helpers import LoggingMixin
from argilla._models import RecordModel
from argilla._exceptions import RecordsIngestionError
from argilla.client import Argilla
from argilla.records._io import GenericIO, HFDataset, HFDatasetsIO, JsonIO
from argilla.records._mapping import IngestedRecordMapper
from argilla.records._resource import Record
from argilla.records._search import Query

if TYPE_CHECKING:
    from argilla.datasets import Dataset


class DatasetRecordsIterator:
    """This class is used to iterate over records in a dataset"""

    def __init__(
        self,
        dataset: "Dataset",
        client: "Argilla",
        query: Optional[Query] = None,
        start_offset: int = 0,
        batch_size: Optional[int] = None,
        with_suggestions: bool = False,
        with_responses: bool = False,
        with_vectors: Optional[Union[str, List[str], bool]] = None,
    ):
        self.__dataset = dataset
        self.__client = client
        self.__query = query or Query()
        self.__offset = start_offset or 0
        self.__batch_size = batch_size or 100
        self.__with_suggestions = with_suggestions
        self.__with_responses = with_responses
        self.__with_vectors = with_vectors
        self.__records_batch = []

    def __iter__(self):
        return self

    def __next__(self) -> Record:
        if not self._has_local_records():
            self._fetch_next_batch()
            if not self._has_local_records():
                raise StopIteration()
        return self._next_record()

    def _next_record(self) -> Record:
        return self.__records_batch.pop(0)

    def _has_local_records(self) -> bool:
        return len(self.__records_batch) > 0

    def _fetch_next_batch(self) -> None:
        self.__records_batch = list(self._list())
        self.__offset += len(self.__records_batch)

    def _list(self) -> Sequence[Record]:
        for record_model in self._fetch_from_server():
            yield Record.from_model(model=record_model, dataset=self.__dataset)

    def _fetch_from_server(self) -> List[RecordModel]:
        if not self.__dataset.exists():
            raise ValueError(f"Dataset {self.__dataset.name} does not exist on the server.")
        if self._is_search_query():
            return self._fetch_from_server_with_search()
        return self._fetch_from_server_with_list()

    def _fetch_from_server_with_list(self) -> List[RecordModel]:
        return self.__client.api.records.list(
            dataset_id=self.__dataset.id,
            limit=self.__batch_size,
            offset=self.__offset,
            with_responses=self.__with_responses,
            with_suggestions=self.__with_suggestions,
            with_vectors=self.__with_vectors,
        )

    def _fetch_from_server_with_search(self) -> List[RecordModel]:
        search_items, total = self.__client.api.records.search(
            dataset_id=self.__dataset.id,
            query=self.__query.api_model(),
            limit=self.__batch_size,
            offset=self.__offset,
            with_responses=self.__with_responses,
            with_suggestions=self.__with_suggestions,
        )
        return [record_model for record_model, _ in search_items]

    def _is_search_query(self) -> bool:
        return bool(self.__query and (self.__query.query or self.__query.filter))

    def to_list(self, flatten: bool) -> List[Dict[str, Any]]:
        return GenericIO.to_list(records=list(self), flatten=flatten)

    def to_dict(self, flatten: bool, orient: str) -> Dict[str, Any]:
        data = GenericIO.to_dict(records=list(self), flatten=flatten, orient=orient)
        return data

    def to_json(self, path: Union[Path, str]) -> Path:
        return JsonIO.to_json(records=list(self), path=path)

    def to_datasets(self) -> "HFDataset":
        return HFDatasetsIO.to_datasets(records=list(self))


class DatasetRecords(Iterable[Record], LoggingMixin):
    """This class is used to work with records from a dataset and is accessed via `Dataset.records`.
    The responsibility of this class is to provide an interface to interact with records in a dataset,
    by adding, updating, fetching, querying, deleting, and exporting records.

    Attributes:
        client (Argilla): The Argilla client object.
        dataset (Dataset): The dataset object.
    """

    _api: RecordsAPI

    DEFAULT_BATCH_SIZE = 256

    def __init__(self, client: "Argilla", dataset: "Dataset"):
        """Initializes a DatasetRecords object with a client and a dataset.
        Args:
            client: An Argilla client object.
            dataset: A Dataset object.
        """
        self.__client = client
        self.__dataset = dataset
        self._api = self.__client.api.records

    def __iter__(self):
        return DatasetRecordsIterator(self.__dataset, self.__client, with_suggestions=True, with_responses=True)

    def __call__(
        self,
        query: Optional[Union[str, Query]] = None,
        batch_size: Optional[int] = DEFAULT_BATCH_SIZE,
        start_offset: int = 0,
        with_suggestions: bool = True,
        with_responses: bool = True,
        with_vectors: Optional[Union[List, bool, str]] = None,
    ) -> DatasetRecordsIterator:
        """Returns an iterator over the records in the dataset on the server.

        Parameters:
            query: A string or a Query object to filter the records.
            batch_size: The number of records to fetch in each batch. The default is 256.
            start_offset: The offset from which to start fetching records. The default is 0.
            with_suggestions: Whether to include suggestions in the records. The default is True.
            with_responses: Whether to include responses in the records. The default is True.
            with_vectors: A list of vector names to include in the records. The default is None.
                If a list is provided, only the specified vectors will be included.
                If True is provided, all vectors will be included.

        Returns:
            An iterator over the records in the dataset on the server.

        """
        if query and isinstance(query, str):
            query = Query(query=query)

        if with_vectors:
            self._validate_vector_names(vector_names=with_vectors)

        return DatasetRecordsIterator(
            dataset=self.__dataset,
            client=self.__client,
            query=query,
            batch_size=batch_size,
            start_offset=start_offset,
            with_suggestions=with_suggestions,
            with_responses=with_responses,
            with_vectors=with_vectors,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.__dataset})"

    ############################
    # Public methods
    ############################

    def log(
        self,
        records: Union[List[dict], List[Record], HFDataset],
        mapping: Optional[Dict[str, Union[str, Sequence[str]]]] = None,
        user_id: Optional[UUID] = None,
        batch_size: int = DEFAULT_BATCH_SIZE,
    ) -> "DatasetRecords":
        """Add or update records in a dataset on the server using the provided records.
        If the record includes a known `id` field, the record will be updated.
        If the record does not include a known `id` field, the record will be added as a new record.
        See `rg.Record` for more information on the record definition.

        Parameters:
            records: A list of `Record` objects, a Hugging Face Dataset, or a list of dictionaries representing the records.
                     If records are defined as a dictionaries or a dataset, the keys/ column names should correspond to the
                     fields in the Argilla dataset's fields and questions. `id` should be provided to identify the records when updating.
            mapping: A dictionary that maps the keys/ column names in the records to the fields or questions in the Argilla dataset.
                     To assign an incoming key or column to multiple fields or questions, provide a list or tuple of field or question names.
            user_id: The user id to be associated with the records' response. If not provided, the current user id is used.
            batch_size: The number of records to send in each batch. The default is 256.

        Returns:
            A list of Record objects representing the updated records.
        """
        record_models = self._ingest_records(records=records, mapping=mapping, user_id=user_id or self.__client.me.id)
        batch_size = self._normalize_batch_size(
            batch_size=batch_size,
            records_length=len(record_models),
            max_value=self._api.MAX_RECORDS_PER_UPSERT_BULK,
        )

        created_or_updated = []
        records_updated = 0

        for batch in tqdm(
            iterable=range(0, len(records), batch_size),
            desc="Sending records...",
            total=len(records) // batch_size,
            unit="batch",
        ):
            self._log_message(message=f"Sending records from {batch} to {batch + batch_size}.")
            batch_records = record_models[batch : batch + batch_size]
            models, updated = self._api.bulk_upsert(dataset_id=self.__dataset.id, records=batch_records)
            created_or_updated.extend([Record.from_model(model=model, dataset=self.__dataset) for model in models])
            records_updated += updated

        records_created = len(created_or_updated) - records_updated
        self._log_message(
            message=f"Updated {records_updated} records and added {records_created} records to dataset {self.__dataset.name}",
            level="info",
        )

        return self

    def delete(
        self,
        records: List[Record],
    ) -> List[Record]:
        """Delete records in a dataset on the server using the provided records
            and matching based on the id.

        Parameters:
            records: A list of `Record` objects representing the records to be deleted.

        Returns:
            A list of Record objects representing the deleted records.

        """
        mapping = None
        user_id = self.__client.me.id

        record_models = self._ingest_records(records=records, mapping=mapping, user_id=user_id)

        self._api.delete_many(dataset_id=self.__dataset.id, records=record_models)

        self._log_message(
            message=f"Deleted {len(record_models)} records from dataset {self.__dataset.name}",
            level="info",
        )

        return record_models

    def to_dict(self, flatten: bool = False, orient: str = "names") -> Dict[str, Any]:
        """
        Return the records as a dictionary. This is a convenient shortcut for dataset.records(...).to_dict().

        Parameters:
            flatten (bool): The structure of the exported dictionary.
                - True: The record fields, metadata, suggestions and responses will be flattened.
                - False: The record fields, metadata, suggestions and responses will be nested.
            orient (str): The orientation of the exported dictionary.
                - "names": The keys of the dictionary will be the names of the fields, metadata, suggestions and responses.
                - "index": The keys of the dictionary will be the id of the records.
        Returns:
            A dictionary of records.

        """
        return self().to_dict(flatten=flatten, orient=orient)

    def to_list(self, flatten: bool = False) -> List[Dict[str, Any]]:
        """
        Return the records as a list of dictionaries. This is a convenient shortcut for dataset.records(...).to_list().

        Parameters:
            flatten (bool): The structure of the exported dictionaries in the list.
                - True: The record keys are flattened and a dot notation is used to record attributes and their attributes . For example, `label.suggestion` and `label.response`. Records responses are spread across multiple columns for values and users.
                - False: The record fields, metadata, suggestions and responses will be nested dictionary with keys for record attributes.
        Returns:
            A list of dictionaries of records.
        """
        data = self().to_list(flatten=flatten)
        return data

    def to_json(self, path: Union[Path, str]) -> Path:
        """
        Export the records to a file on disk.

        Parameters:
            path (str): The path to the file to save the records.

        Returns:
            The path to the file where the records were saved.

        """
        return self().to_json(path=path)

    def from_json(self, path: Union[Path, str]) -> List[Record]:
        """Creates a DatasetRecords object from a disk path to a JSON file.
            The JSON file should be defined by `DatasetRecords.to_json`.

        Args:
            path (str): The path to the file containing the records.

        Returns:
            DatasetRecords: The DatasetRecords object created from the disk path.

        """
        records = JsonIO._records_from_json(path=path)
        return self.log(records=records)

    def to_datasets(self) -> HFDataset:
        """
        Export the records to a HFDataset.

        Returns:
            The dataset containing the records.

        """

        return self().to_datasets()

    ############################
    # Private methods
    ############################

    def _ingest_records(
        self,
        records: Union[List[Dict[str, Any]], List[Record], HFDataset],
        mapping: Optional[Dict[str, Union[str, Sequence[str]]]] = None,
        user_id: Optional[UUID] = None,
    ) -> List[RecordModel]:
        """Ingests records from a list of dictionaries, a Hugging Face Dataset, or a list of Record objects."""

        if len(records) == 0:
            raise ValueError("No records provided to ingest.")

        if HFDatasetsIO._is_hf_dataset(dataset=records):
            records = HFDatasetsIO._record_dicts_from_datasets(dataset=records)

        ingested_records = []
        record_mapper = IngestedRecordMapper(mapping=mapping, dataset=self.__dataset, user_id=user_id)
        for record in records:
            try:
                if isinstance(record, dict):
                    record = record_mapper(data=record)
                elif isinstance(record, Record):
                    record.dataset = self.__dataset
                else:
                    raise ValueError(
                        "Records should be a a list Record instances, "
                        "a Hugging Face Dataset, or a list of dictionaries representing the records."
                        f"Found a record of type {type(record)}: {record}."
                    )
            except Exception as e:
                raise RecordsIngestionError(f"Failed to ingest record from dict {record}: {e}")
            ingested_records.append(record.api_model())
        return ingested_records

    def _normalize_batch_size(self, batch_size: int, records_length, max_value: int):
        norm_batch_size = min(batch_size, records_length, max_value)

        if batch_size != norm_batch_size:
            self._log_message(
                message=f"The provided batch size {batch_size} was normalized. Using value {norm_batch_size}.",
                level="warning",
            )

        return norm_batch_size

    def _validate_vector_names(self, vector_names: Union[List[str], str]) -> None:
        if not isinstance(vector_names, list):
            vector_names = [vector_names]
        for vector_name in vector_names:
            if isinstance(vector_name, bool):
                continue
            if vector_name not in self.__dataset.schema:
                raise ValueError(f"Vector field {vector_name} not found in dataset schema.")



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import uuid

import pytest

from argilla import UserResponse, Response, Dataset, Workspace
from argilla._models import UserResponseModel, ResponseStatus


class TestResponses:
    def test_create_user_response(self):
        user_id = uuid.uuid4()
        response = UserResponse(
            responses=[
                Response(question_name="question", value="answer", user_id=user_id),
                Response(question_name="other-question", value="answer", user_id=user_id),
            ],
        )

        assert response.to_dict() == {
            "values": {
                "question": {"value": "answer"},
                "other-question": {"value": "answer"},
            },
            "status": "draft",
            "user_id": str(user_id),
        }

    def test_create_submitted_user_responses(self):
        user_id = uuid.uuid4()
        response = UserResponse(
            responses=[
                Response(question_name="question", value="answer", user_id=user_id, status="submitted"),
                Response(question_name="other-question", value="answer", user_id=user_id, status="submitted"),
            ],
        )

        assert response.to_dict() == {
            "values": {
                "question": {"value": "answer"},
                "other-question": {"value": "answer"},
            },
            "status": "submitted",
            "user_id": str(user_id),
        }

    def test_create_user_response_with_multiple_status(self):
        user_id = uuid.uuid4()
        response = UserResponse(
            responses=[
                Response(question_name="question", value="answer", user_id=user_id, status="draft"),
                Response(question_name="other-question", value="answer", user_id=user_id, status="submitted"),
            ],
        )

        assert response.to_dict() == {
            "values": {
                "question": {"value": "answer"},
                "other-question": {"value": "answer"},
            },
            "status": "draft",
            "user_id": str(user_id),
        }

    def test_create_user_response_with_multiple_user_id(self):
        user_id = uuid.uuid4()
        other_user_id = uuid.uuid4()

        with pytest.raises(ValueError, match="Multiple user_ids found in user responses"):
            UserResponse(
                responses=[
                    Response(question_name="question", value="answer", user_id=user_id),
                    Response(question_name="other-question", value="answer", user_id=other_user_id),
                ],
            )

    def test_create_user_response_from_draft_response_model_without_values(self):
        model = UserResponseModel(values={}, status=ResponseStatus.draft, user=uuid.uuid4())
        response = UserResponse.from_model(
            model=model, dataset=Dataset(name="burr", workspace=Workspace(name="test", id=uuid.uuid4()))
        )
        assert len(response.responses) == 0
        assert response.user_id is None
        assert response.status == ResponseStatus.draft



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from rich import print

import logging

LOG_LEVEL_MAP = {
    "debug": logging.DEBUG,
    "info": logging.INFO,
    "warning": logging.WARNING,
    "error": logging.ERROR,
    "critical": logging.CRITICAL,
}

logger = logging.getLogger(name="argilla.sdk")


def log_message(message: str, level: str = "info") -> None:
    """Log a message at the specified level.
    Args:
        message (str): The message to log.
        level (str): The log level to use. Defaults to "info".
    """
    level_int = LOG_LEVEL_MAP.get(level, logging.INFO)
    logger.log(level=level_int, msg=message)


def log_interactive(message: str) -> None:
    """Log a message to the console in an interactive environment.
    Args:
        message (str): The message to log.
    """
    print(message)


class LoggingMixin:
    """A utility mixin for logging from a `Resource` class."""

    def _log_message(self, message: str, level: str = "info") -> None:
        class_name = self.__class__.__name__
        message = f"{class_name}: {message}"
        if self._is_interactive() and level != "info":
            log_interactive(message=message)
        else:
            log_message(level=level, message=message)

    def _is_interactive(self) -> bool:
        import __main__ as main

        return not hasattr(main, "__file__")



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Dict, List, Union, Optional, Type, TYPE_CHECKING, Any

from argilla.records._io._generic import GenericIO

if TYPE_CHECKING:
    from argilla.records import Record


def _resolve_hf_datasets_type() -> Optional[Type["HFDataset"]]:
    """This function resolves the `datasets.Dataset` type safely in case the datasets package is not installed.

    Returns:
        Optional[Type]: The Dataset class definition in case the datasets package is installed. Otherwise, None.
    """
    try:
        from datasets import Dataset

        return Dataset
    except ImportError:
        return None


HFDataset = _resolve_hf_datasets_type()


class HFDatasetsIO:
    @staticmethod
    def _is_hf_dataset(dataset: Any) -> bool:
        """Check if the object is a Hugging Face dataset.

        Parameters:
            dataset (Dataset): The object to check.

        Returns:
            bool: True if the object is a Hugging Face dataset, False otherwise.
        """
        datasets_class = _resolve_hf_datasets_type()
        if datasets_class is None:
            return False
        return isinstance(dataset, datasets_class)

    @staticmethod
    def to_datasets(records: List["Record"]) -> "HFDataset":
        """
        Export the records to a Hugging Face dataset.

        Returns:
            The dataset containing the records.
        """
        dataset_class = _resolve_hf_datasets_type()
        if dataset_class is None:
            raise ImportError("Hugging Face datasets is not installed. Please install it using `pip install datasets`.")
        record_dicts = GenericIO.to_dict(records, flatten=True)
        dataset = dataset_class.from_dict(record_dicts)
        return dataset

    @staticmethod
    def _record_dicts_from_datasets(dataset: "HFDataset") -> List[Dict[str, Union[str, float, int, list]]]:
        """Creates a dictionaries from a HF dataset that can be passed to DatasetRecords.add or DatasetRecords.update.

        Parameters:
            dataset (Dataset): The dataset containing the records.

        Returns:
            Generator[Dict[str, Union[str, float, int, list]], None, None]: A generator of dictionaries to be passed to DatasetRecords.add or DatasetRecords.update.
        """
        record_dicts = []
        for example in dataset.to_iterable_dataset():
            record_dicts.append(example)
        return record_dicts



# Copyright 2024-present, Argilla, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import warnings

try:
    from argilla_v1 import *  # noqa
except ModuleNotFoundError as ex:
    raise Exception(
        'The package argilla-v1 is not installed. Please install it by typing: pip install "argilla[legacy]"',
    ) from ex


def deprecation(message: str):
    warnings.warn(message, DeprecationWarning, stacklevel=2)


deprecation(
    "The module `argilla_sdk.v1` has been include for migration purposes. "
    "It's deprecated and will be removed in the future."
)
